"""
Trains a neural network to fit the opacity data.
"""
import pickle

import keras
import numpy as np
import tensorflow as tf
from tensorflow.keras import layers

from cortecs.fit.metric_plot import *


# todo: loop over all wavelengths.
def prep_neural_net(cross_section):
    pass


def unravel_data(x, y, z=None, tileboth=False):
    """
    unravels the data into a single column. takes log as well.

    todo: move to a utils file?

    Inputs
    ------
    x: array-like
        first dimension of data.
    y: array-like
        second dimension of data.
    z: array-like
        third dimension of data.
    """

    if type(z) == type(None):
        if tileboth:
            return np.tile(np.log10(x), len(y))
        return np.repeat(np.log10(x), len(y))
    if tileboth:
        return np.tile(np.tile(np.log10(x), len(y)), len(z))

    return np.tile(np.repeat(np.log10(x), len(y)), len(z))


def fit_neural_net(
    cross_section,
    P,
    T,
    wl,
    all_wl=False,
    n_layers=3,
    n_neurons=8,
    activation="sigmoid",
    learn_rate=0.04,
    loss="mean_squared_error",
    epochs=2000,
    verbose=0,
    sequential_model=None,
    plot=False,
):
    """
    trains a neural network to fit the opacity data.
    :param Opac: not an actual opac.
    :param n_layers:
    :param n_neurons:
    :param activation:
    :param learn_rate:
    :param loss:
    :param epochs:
    :param verbose:
    :param sequential_model: if not None (and instead a sequential object), this overwrites the other neural net
                            parameters.
    :return:
        :history: the history object from the keras fit method
        :neural_network: the trained neural network
    """
    if sequential_model is None:
        layers_list = []
        for i in range(n_layers):
            layers_list += [layers.Dense(n_neurons, activation=activation)]

        layers_list += [layers.Dense(1)]  # final layer to predict single value

        neural_network = keras.Sequential(layers_list)

    else:
        neural_network = sequential_model

    # running the legacy Adam optimizer for the Mac folks!
    neural_network.compile(
        loss=loss, optimizer=tf.keras.optimizers.legacy.Adam(learn_rate)
    )

    # unpack opacity data. todo: not repeat this for every wavelength!
    if all_wl:
        raise NotImplementedError("todo: implement all_wl")
    P_unraveled = unravel_data(P, T, wl)
    T_unraveled = unravel_data(T, P, wl, tileboth=True)

    # todo: try to predict everything in one big net, not wavelength-independent?
    input_array = np.column_stack([P_unraveled, T_unraveled])
    predict_data_flattened = cross_section.flatten()

    history = neural_network.fit(
        input_array,
        predict_data_flattened,  # totally fine to overfit! we're not extrapolating, nor are we interpolating.
        verbose=verbose,
        epochs=epochs,
    )

    if plot:
        plot_loss(history)

    return history, neural_network


def save_neural_network(neural_network, filename):
    """
    saves the neural network to a file.
    :param neural_network:
    :param filename:
    :return:
    """
    n_layers = len(neural_network.layers)
    all_weights = []
    all_biases = []

    for layer in range(n_layers):
        layer_weights = neural_network.layers[layer].get_weights()[0]
        layer_biases = neural_network.layers[layer].get_weights()[1]
        all_weights += [layer_weights]
        all_biases += [layer_biases]

    # save all_weights and all_biases to a pickle
    with open(filename, "wb") as f:
        pickle.dump([all_weights, all_biases], f)

    # now, let's read it back in and make sure it's the same.

    return



"""
Reads opacity data from various sources.



author: @arjunsavel
"""
import pickle

import h5py
import numpy as np
from tqdm import tqdm

AMU = 1.6605390666e-24  # atomic mass unit in cgs. From astropy!


class loader_base(object):
    """
    loads in opacity data from various sources. To be passed on to Opac object.
    """

    wl_key = "wno"
    T_key = "T"
    P_key = "P"
    cross_section_key = "xsec"
    wl_style = "wno"

    def __init__(self):
        """
        nothing to do here
        """
        pass

    def load(self, filename):
        """
        loads in opacity data from various sources. To be passed on to Opac object.

        Inputs
        ------
        filename : str
            name of file to load

        Outputs
        -------
        wl : np.ndarray
            wavelengths
        T : np.ndarray
            temperatures
        P : np.ndarray
            pressures
        cross_section : np.ndarray
            cross sections for the opacity
        """

        hf = h5py.File(filename, "r")
        wl = np.array(hf[self.wl_key], dtype=np.float64)
        T = np.array(hf[self.T_key], dtype=np.float64)
        P = np.array(hf[self.P_key], dtype=np.float64)
        cross_section = np.array(hf[self.cross_section_key], dtype=np.float64)
        hf.close()

        if self.wl_style == "wno":
            wl = 1e4 / wl
        return wl, T, P, cross_section


class loader_helios(loader_base):
    """
    loads in opacity data that are produced with the HELIOS ktable function.
    """

    wl_key = "wavelengths"
    T_key = "temperatures"
    P_key = "pressures"
    cross_section_key = "opacities"
    wl_style = "wl"


class loader_platon(loader_base):
    """
    loads in opacity data that are used with the PLATON code.
    """

    # atomic weights of molecules
    species_weight_dict = {
        "CO": 28.01,
        "H2O": 18.015,
        "CH4": 16.04,
        "NH3": 17.031,
        "CO2": 44.009,
        "HCN": 27.026,
        "C2H2": 26.038,
        "H2S": 34.08,
        "PH3": 33.997,
        "VO": 66.94,
        "TiO": 63.866,
        "Na": 22.99,
        "K": 39.098,
        "FeH": 55.845,
        "H2": 2.016,
        "He": 4.003,
        "H-": 1.008,
        "H": 1.008,
        "He+": 4.003,
        "H+": 1.008,
        "e-": 0.00054858,
        "H2+": 2.016,
        "H2-": 2.016,
        "H3+": 3.024,
        "H3-": 3.024,
        "H2O+": 18.015,
        "H2O-": 18.015,
        "CH4+": 16.04,
        "CH4-": 16.04,
        "C2H4": 28.054,
    }

    def load(self, cross_sec_filename, T_filename="", P_filename="", wl_filename=""):
        """
        loads in opacity data that's built for PLATON. To be passed on to Opac object.

        The temperature grid, pressure grid, and wavelength grid are saved as separate files for PLATON.

        Inputs
        ------
        filename : str
            name of file to load
        cross_sec_filename : str
            name of cross section file
        T_filename : str
            name of temperature file
        P_filename : str
            name of pressure file
        wl_filename : str
            name of wavelength file
        """
        # todo: check wl units. They're in meters here.
        # temperatures are in K.
        # pressures are in Pa, I believe.
        wl = np.load(wl_filename)
        T = np.load(T_filename)
        P = np.load(P_filename)
        cross_section = np.load(
            cross_sec_filename
        )  # packaged as T x P x wl. todo: check packing
        # cross-section units for fitting...? keep the same internally.
        species = cross_sec_filename.split("_")[-1].split(".")[0]
        try:
            self.species_weight = self.species_weight_dict[species]
        except:
            raise KeyError(
                f"Species {species} read from filename and not found in species_weight_dict. Please add it."
            )
        cross_section = cross_section * AMU * self.species_weight * 1e-4
        # and now make it log10
        cross_section = np.log10(cross_section)

        # and set the infs to -104
        cross_section[~np.isfinite(cross_section)] = -104.0

        return wl, T, P, cross_section


class loader_platon_cia(loader_base):
    """
    loads in opacity data that are used with the PLATON code's collision-induced absorption..
    """

    def load(self, cross_sec_filename, T_filename, wl_filename, species_name):
        """
        loads in opacity data that's built for PLATON. To be passed on to Opac object.

        The temperature grid, pressure grid, and wavelength grid are saved as separate files for PLATON.

        Inputs
        ------
        filename : str
            name of file to load
        cross_sec_filename : str
            name of cross section file
        T_filename : str
            name of temperature file
        wl_filename : str
            name of wavelength file
        species_name : tuples
            name of two colliding species. E.g., ('H2', 'CH4'). todo: all at once?
        """
        # todo: check wl units. They're in meters here.
        # temperatures are in K.
        # pressures are in Pa, I believe.
        wl = np.load(wl_filename)
        T = np.load(T_filename)
        data = pickle.load(
            open(cross_sec_filename, "rb"), encoding="latin1"
        )  # packaged as T x P x wl. todo: check packing
        cross_section = data[species_name]
        return wl, T, cross_section


class loader_exotransmit(loader_base):
    """
    loads in opacity data in the exo-transmit format. This format is also used by PLATON, I believe.
    """

    def get_t_p(self, file):
        """
        Gets the temperatures and pressures of a file from its header.

        Inputs:
            :file: (str) path to file whose header we want.

        Outputs:
            header of file (string)
        """

        f = open(file)
        f1 = f.readlines()
        f.close()

        t_line = f1[0]
        p_line = f1[1]

        temperature_strings = t_line.split(" ")[:-1]
        T = np.array([eval(temp) for temp in temperature_strings])

        pressure_strings = p_line.split(" ")[:-1]
        P = np.array([eval(pres) for pres in pressure_strings])

        del f1
        return T, P

    def get_lams_and_opacities(self, file):
        """
        Takes in an opacity file and returns an array of all wavelengths within the file.

        Returns the opacities, as well — so that the file is only iterated through once.

        Inputs:
            :file: (str) path to opacity file.

        Outputs:
            :wavelengths: (numpy.array) individual wavelength points within the opacity file [m]
        """
        f = open(file)
        f1 = f.readlines()

        wavelengths = []
        opacities = []

        # read through all lines in the opacity file
        for x in tqdm(f1, desc="reading wavelengths"):
            # check if blank line
            if not x:
                continue
            # check if a wavelength line
            commad = x.replace(" ", ",")
            if len(np.array([eval(commad)]).flatten()) == 1:
                wavelengths += [eval(x[:-1])]
            else:
                # the first entry in each opacity line is the pressure
                opacity_string = commad.split(" ")[1:]
                opacity_vals = np.array([eval(opacity) for opacity in opacity_string])
                opacities += [opacity_vals]

        f.close()
        del f1
        return np.array(wavelengths), np.array(opacities)

    def load(self, filename):
        wl, opacities = self.get_lams_and_opacities(filename)
        T, P = self.get_t_p(filename)

        return wl, T, P, opacities



import jax
import jax.numpy as jnp
import numpy as np

# todo: loop over all wavelengths.


def prep_polynomial(cross_section, **kargs):
    pass


# @jax.jit
def fit_polynomial(Z, P, T, prep_res, plot=False, save=False):
    """
    todo: actually take in the Opac object.
    fits a polynomial to the opacity data.
    :param wavelength_ind:
    :param plot:
    :param species:
    :param save:
    :return:
    """
    X, Y = jnp.meshgrid(jnp.log10(T), jnp.log10(P), copy=True)

    X = X.flatten()
    Y = Y.flatten()

    # took out the y**2 term

    A = jnp.array(
        [
            X * 0 + 1,
            X,
            Y,
            X**2,
            X**2 * Y,
            X**2 * Y**2,
            X * Y**2,
            X * Y,
            X * Y**3,
            Y**4,
            Y**5,
            Y**6,
            Y**7,
            Y**12,
            X * Y**9,
            np.power(X, 1 / 4),
        ]
    ).T
    B = Z.flatten()

    coeff, r, rank, s = np.linalg.lstsq(A, B, rcond=-1)

    return coeff



"""
Plots histograms and stuff after the fitting has occurred so that the level of fit can be assessed.
"""
import matplotlib.pyplot as plt
import numpy as np


def plot_loss(history):
    """
    plots the loss from the history object. need to substantially rewrite
    TODO: rcparams file to make things look nice, fontwise and such?
    :param history:
    :return:
    """
    plt.plot(history.history["loss"], label="loss", color="teal", lw=3)
    plt.xlabel("Epoch", fontsize=20)
    plt.ylabel("Error", fontsize=20)  # need the units on this
    plt.yscale("log")
    plt.legend()


def plot_quantity(species_quantity_dict, quantity="mean", log=True, nbins=20):
    """
    plots the quantities related to the species fitting.

    Inputs
    ------
    species_quantity_dict: dict
        dictionary of species and the quantities associated with them.
    quantity: str
        quantity to plot. 'mean', 'std', 'min', 'max', 'median', 'percentile'
    log: bool
        whether to plot the log of the quantity.
    nbins: int
        number of bins to use in the histogram.

    """
    # todo: colormaps

    # find the bins
    min = np.inf
    max = -np.inf
    for species in species_quantity_dict.keys():
        species_min = np.min(species_quantity_dict[species])
        min = np.min([min, species_min])

        species_max = np.max(species_quantity_dict[species])
        max = np.max([min, species_max])

    if log:
        bins = np.logspace(np.log10(min), np.log10(max), nbins)
    else:
        bins = np.linspace(min, max, nbins)

    for species in species_quantity_dict.keys():
        if log:
            quantity = np.log10(species_quantity_dict[species])
        else:
            quantity = species_quantity_dict[species]

        plt.hist(quantity, label=species, lw=5, histtype="step", bins=bins)

    plt.legend()
    plt.title("Fitting quality: " + quantity)



"""
Functions that can be used to interpolate CIA to a higher
resolution and subsequently chunk it up.
Example instructions / workflow:
>>> reference_file = '../opacFe/opacFe.dat'
>>> CIA_file = 'opacCIA.dat'
>>> interpolate_CIA(CIA_file, reference_file)
>>> CIA_file = 'opacCIA_highres.dat'
>>> ref_file_base = '../opacFe/opacFe'
>>> chunk_wavelengths_CIA(CIA_file, ref_file_base)
And that should work!

todo: generalize to PLATON CIA.

author: @arjunsavel
"""
import os
from glob import glob

import numpy as np
import pandas as pd
from tqdm import tqdm

from cortecs.opac.chunking import *

######################### Pt. 1: Interpolation ####################################


def interpolate_CIA(CIA_file, reference_file):
    """
    Interpolates a CIA file to a higher resolution, using the wavelength grid
    of a reference file. Note: This function assumes that the CIA file has
    Hels, HeHs, CH4CH4s, H2Hes, H2CH4s, H2Hs, H2H2s, and CO2CO2s.
    Inputs
    ------
        :CIA_file: (str) path to CIA file to be interpolated. e.g.,
                    'opacCIA/opacCIA.dat'
        :reference_file: (str) path to opacity file with the wavelength grid of interest. e.g.,
                    'opacFe/opacFe.dat'
    Outputs
    -------
        None
    Side effects
    -------------
        Creates a file with "hires" attached to the end of CIA_file that has been interpolated
        to higher resolution.
    """

    real_wavelength_grid = get_wav_grid(reference_file, progress=True)

    f = open(CIA_file)
    f1 = f.readlines()
    f.close()

    temperatures = []
    wavelengths = []
    # todo: refactor. has to be a cleaner way to do this! infer the columns, etc.
    Hels = []
    HeHs = []
    CH4CH4s = []
    H2Hes = []
    H2CH4s = []
    H2Hs = []
    H2H2s = []
    CO2CO2s = []

    # read through all lines in the CIA file
    for line in tqdm(f1[1:], desc="Reading CIA file"):
        if not line or line == "\n":
            continue  # don't want it to break!
        if len(line.split(" ")) == 1 and line != "\n":  # this is a new temperature
            temperature = eval(line[:-1])
            continue  # nothing else on this line
        values = [eval(value) for value in line.split(" ")[::3][:-1]]
        temperatures += [temperature]
        wavelengths += [values[0]]
        Hels += [values[1]]
        HeHs += [values[2]]
        CH4CH4s += [values[3]]
        H2Hes += [values[4]]
        H2CH4s += [values[5]]
        H2Hs += [values[6]]
        H2H2s += [values[7]]
        CO2CO2s += [values[8]]

    # easier to slice with a dataframe later on
    df = pd.DataFrame(
        {
            "temp": temperatures,
            "wav": wavelengths,
            "Hel": Hels,
            "HeH": HeHs,
            "CH4CH4": CH4CH4s,
            "H2He": H2Hes,
            "H2CH4": H2CH4s,
            "H2H": H2Hs,
            "H2H2": H2H2s,
            "CO2CO2": CO2CO2s,
        }
    )

    # perform interpolation

    interped_Hels = []
    interped_HeHs = []
    interped_CH4CH4s = []
    interped_H2Hes = []
    interped_H2CH4s = []
    interped_H2Hs = []
    interped_H2H2s = []
    interped_CO2CO2s = []
    interped_wavelengths = []
    interped_temps = []

    for unique_temp in tqdm(df.temp.unique()):
        sub_df = df[df.temp == unique_temp]

        # convert to lists so that they're neatly nested lists
        interped_Hels += list(np.interp(real_wavelength_grid, sub_df.wav, sub_df.Hel))
        interped_HeHs += list(np.interp(real_wavelength_grid, sub_df.wav, sub_df.HeH))
        interped_CH4CH4s += list(
            np.interp(real_wavelength_grid, sub_df.wav, sub_df.CH4CH4)
        )
        interped_H2Hes += list(np.interp(real_wavelength_grid, sub_df.wav, sub_df.H2He))
        interped_H2CH4s += list(
            np.interp(real_wavelength_grid, sub_df.wav, sub_df.H2CH4)
        )
        interped_H2Hs += list(np.interp(real_wavelength_grid, sub_df.wav, sub_df.H2H))
        interped_H2H2s += list(np.interp(real_wavelength_grid, sub_df.wav, sub_df.H2H2))
        interped_CO2CO2s += list(
            np.interp(real_wavelength_grid, sub_df.wav, sub_df.CO2CO2)
        )

        interped_wavelengths += real_wavelength_grid
        interped_temps += list(np.ones_like(real_wavelength_grid) * unique_temp)

    # write to a new file

    new_string = []

    buffer = "   "  # there's a set of spaces between each string!
    for i in tqdm(range(len(interped_Hels))):
        # the first line gets different treatment!
        if i == 0:
            temp = 500
            new_string += ["{:.12e}".format(temp) + "\n"]
        if interped_temps[i] != temp:
            temp = interped_temps[i]
            new_string += ["{:.12e}".format(temp) + "\n"]
        wavelength = "{:.12e}".format(interped_wavelengths[i])
        Hel_string = "{:.12e}".format(interped_Hels[i])
        HeH_string = "{:.12e}".format(interped_HeHs[i])
        CH4CH4_string = "{:.12e}".format(interped_CH4CH4s[i])
        H2He_string = "{:.12e}".format(interped_H2Hes[i])
        H2CH4_string = "{:.12e}".format(interped_H2CH4s[i])
        H2H_string = "{:.12e}".format(interped_H2Hs[i])
        H2H2_string = "{:.12e}".format(interped_H2H2s[i])
        CO2CO2_string = "{:.12e}".format(interped_CO2CO2s[i])
        new_string += [
            wavelength
            + buffer
            + Hel_string
            + buffer
            + HeH_string
            + buffer
            + CH4CH4_string
            + buffer
            + H2He_string
            + buffer
            + H2CH4_string
            + buffer
            + H2H_string
            + buffer
            + H2H2_string
            + buffer
            + CO2CO2_string
            + buffer
            + "\n"
        ]

    # todo: check the insert. and can pull wavelength grid.
    new_string.insert(
        0,
        "500.000 600.000 700.000 800.000 900.000 1000.000 1100.000 1200.000 1300.000 1400.000 1500.000 1600.000 1700.000 1800.000 1900.000 2000.000 2100.000 2200.000 2300.000 2400.000 2500.000 2600.000 2700.000 2800.000 2900.000 3000.000 3100.000 3200.000 3300.000 3400.000 3500.000 3600.000 3700.000 3800.000 3900.000 4000.000 4100.000 4200.000 4300.000 4400.000 4500.000 4600.000 4700.000 4800.000 4900.000 5000.000 \n",
    )
    new_file = CIA_file.split(".dat")[0]
    new_file += "_highres.dat"
    f2 = open(new_file, "w")
    f2.writelines(new_string)
    f2.close()

    return


def get_wav_grid(file, progress=False):
    """
    Returns the wavelength grid used in an opacity file.
    Inputs
    -------
        :file: (str) path to opacity file with the wavelength grid of interest. e.g.,
                    'opacFe/opacFe.dat'
        :progress: (bool) whether or not to include a progress bar (if tqdm is installed).
                    Useful for the biggest opacity file!
    Outputs
    -------
        :wav_grid: (np.array) wavelength values for which the opacity file had been
                    computed.
    """

    wav_grid = []

    f = open(file)
    f1 = f.readlines()
    f.close()

    if progress:
        iterator = tqdm(f1[2:], desc="Grabbing wavelength grid")
    else:
        iterator = f1[2:]

    # read through all lines in the opacity file; first few lines are header!
    for x in iterator:
        # skip blank lines
        if not x:
            continue
        commad = x.replace(" ", ",")

        # check if a wavelength line
        if len(np.array([eval(commad)]).flatten()) == 1:
            wav_grid += [np.array([eval(commad)]).flatten()[0]]

    return wav_grid


######################### Pt. 2: Chunking ####################################


def chunk_wavelengths_CIA(file, ref_file_base, numfiles):
    """
    Performs chunking based on the reference file's wavelength chunking.
    Inputs
    -------
        :file: (str) path to CIA file that should be chunked. e.g., opacCIA_highres.dat.
        :ref_file_base: (str) path base to a set of reference files that are already chunked
                        on the desired wavelength grid. e.g., ../opacFe/opacFe
    """

    header = get_header(file)

    # now get chunks

    f = open(file)
    f1 = f.readlines()
    f.close()

    ticker = 0
    file_suffix = 0

    # determine how many wavelengths are in a given chunk
    if "chunk_list.txt" not in os.listdir():
        chunk_list = []
        for i in tqdm(range(numfiles), desc="Putting together chunk list"):
            chunk_list += [get_wav_per_chunk(i, ref_file_base)]

        np.savetxt("chunk_list.txt", chunk_list)

    else:
        chunk_list = np.loadtxt("chunk_list.txt")

    wav_per_chunk = chunk_list[0]  # to start

    ntemps = 0

    for line in tqdm(
        f1, desc="Writing lines to chunk files"
    ):  # read through all lines in the opacity file
        if not line or line == "\n":
            continue  # don't want it to break

        if len(line.split(" ")) == 1:  # this is a new temperature
            temperature = line
            ntemps += 1
            continue  # nothing else on this line
        ticker += 1
        if ticker == wav_per_chunk:
            file_suffix += 1  # start writing to different file
            ticker = 0
            true_file_suffix = (file_suffix) % numfiles
            wav_per_chunk = chunk_list[true_file_suffix]

            write_to_file(temperature, file, file_suffix, ntemps, numfiles)
        write_to_file(line, file, file_suffix, ntemps, numfiles)  # just writes line

    true_header = header[0]

    for chunk_file in glob(file[:-4] + "*"):
        if chunk_file != file:
            prepend_line(chunk_file, true_header)
    return


def get_wav_per_chunk(file_suffix, ref_file_base):
    """
    Grabs the number of wavelengths of a given chunk.
    Inputs
    ------
        :file_suffix: (int) number corresponding to the given chunk. e.g., 1.
        :ref_file_base: (str) path base to a set of reference files that are already chunked
                        on the desired wavelength grid. e.g., ../opacFe/opacFe
    Outputs
    --------
        :len_grid: (int) number of wavelengths in chunk.
    """

    file = ref_file_base + str(file_suffix) + ".dat"

    real_wavelength_grid = get_wav_grid(file)

    len_grid = len(real_wavelength_grid)

    return len_grid


def prepend_line(file_name, line):
    """
    Insert given string as a new line at the beginning of a file. Used to fix the header within
    the chunked CIA files.
    """

    # define name of temporary dummy file
    dummy_file = file_name + ".bak"
    # open original file in read mode and dummy file in write mode
    with open(file_name, "r") as read_obj, open(dummy_file, "w") as write_obj:
        # Write given line to the dummy file
        write_obj.write(line + "\n")
        # Read lines from original file one by one and append them to the dummy file
        for line in read_obj:
            write_obj.write(line)

    # remove original file
    os.remove(file_name)
    # Rename dummy file as the original file
    os.rename(dummy_file, file_name)


def write_to_file(line, file, file_suffix, ntemps, numfiles):
    """
    Writes a line to a file.
    Inputs
    ------
        :line: (str) line to be written.
        :file: (str) path to file being chunked. e.g., opacCIA_highres.dat
        :file_suffix: (int) number of chunk.
        :ntemps: (int) number of temperatures in grid.
    Outputs
    --------
        None
    Side effects
    -------------
        Writes a line to a file!
    """
    true_file_suffix = (file_suffix) % numfiles
    true_filename = f"{file[:-4] + str(true_file_suffix) + '.dat'}"
    f = open(true_filename, "a")
    f.write(line)
    f.close()



import warnings

import numpy as np
from tqdm import tqdm

# todo: loop over all wavelengths.


def standardize_cube(flux_cube):
    """
    Prepare a flux cube for PCA.
    """
    nf, nx = flux_cube.shape

    """ Looping over orders """
    mat = flux_cube.copy()

    """ Looping over columns / pixels """
    if np.any(np.isnan(mat)):
        print("bad!")
    for i in range(nf):
        mat[i, :] -= np.mean(mat[i, :])
        mat[i, :] /= np.std(mat[i, :])
    """ Updating the output cube and plotting """
    fStd = mat

    return fStd


def do_svd(fStd, nc, nx):
    """
    Does the SVD on the standardized flux cube.

    :param fStd:  standardized flux cube
    :param nc:  number of components to keep
    :param nx:  number of pixels
    :return:  xMat, s, vh, u
    """

    xMat = np.ones((nx, nc + 1))  # Note the extra column of 1s
    mat = np.moveaxis(fStd.copy(), 0, -1)
    u, s, vh = np.linalg.svd(mat, full_matrices=False)
    xMat[:, 1:] = u[:, 0:nc]  # The first column is left as a column of ones

    return xMat, s, vh, u


def fit_mlr(cube, xMat):
    """
    Fits the MLR to the flux cube with the PCA outputs and normalizes accordingly.
    """
    X = xMat
    XT = X.T
    Y = np.moveaxis(cube.copy(), 0, -1)
    term1 = np.linalg.inv(np.dot(XT, X))
    term2 = np.dot(term1, XT)
    beta = np.dot(term2, Y)
    fNorm = Y / np.dot(X, beta)
    return fNorm, beta


def do_pca(cube, nc=3):
    """
    Does all the PCA steps. right now for a single spectrum.

    Inputs
    -------
        :wav_for_pca: (n_exp x pixels) it's an array.
        :flux_for_pca: it's the same.
        :nc: (int) number of PCA components to keep.

    """

    # do the PCA
    fStd = standardize_cube(cube)

    nf, nx = cube.shape
    """ Choosing the number of components and initialising the matrix of eigenvectors """

    try:
        xMat, s, vh, u = do_svd(fStd, nc, nx)

        fNorm, beta = fit_mlr(cube, xMat)
    except np.linalg.LinAlgError:
        print("SVD did not converge.")
        return

    return xMat, fStd, fNorm, beta, s, vh, u


def fit_pca(cross_section, P, T, xMat, nc=3, wav_ind=1):
    """
    Fits the PCA to the opacity data.

    Inputs
    -------
        :cross_section: (n_exp x pixels) it's an array.
        :P: pressure grid
        :T: temperature grid
        :nc: (int) number of PCA components to keep.
        :wav_ind: (int) index of wavelength to fit.

    Returns
    -------
        :xMat: (n_exp x nc) PCA components
        :beta: (nc x pixels) PCA coefficients
    """

    _, beta = fit_mlr(cross_section, xMat)
    return beta


def prep_pca(cross_section, wav_ind=-1, nc=2):
    """
    Prepares the opacity data for PCA.

    Inputs
    -------
        :cross_section: (n_exp x pixels) it's an array.
        :wav_ind: (int) index of wavelength to fit.
        :nc: (int) number of PCA components to keep.

    Returns
    -------
        :xMat: (n_exp x nc) PCA components
    """
    single_pres_single_temp = cross_section[:, :, wav_ind]
    if np.all(single_pres_single_temp == single_pres_single_temp[0, 0]):
        raise ValueError(
            "all values are the same at this wavelength index! Try a different one."
        )

    xMat, fStd, fNorm, beta, s, vh, u = do_pca(single_pres_single_temp, nc)
    return xMat


if __name__ == "__main__":
    # redo for the PCA. this is the big one!!
    species_inds = {
        "C2H2": 200,
        "CO_hitemp": 169565,
        "HCN": 200,
        "NH3": 200,
        "H2S": 200,
        "H2O": 200,
        "CH4": 200,
        "H2H2": 200,
        "H2He": 200,
        "13CO": 169565,
    }
    with warnings.catch_warnings():
        species_means_dict = {}
        for species in ["CO_hitemp"]:
            # '13CO', 'HCN', 'NH3', 'H2S', 'H2O', 'CH4',  'H2H2', 'H2He',  'C2H2',]:

            full_opac = species_dict[species]  # todo: fix
            wav_ind = species_inds[species]
            species_means = []

            single_pres_single_temp = full_opac[:, :, wav_ind]

            xMat, fStd, fNorm, beta, s, vh, u = do_pca(single_pres_single_temp, 2)

            betas = []

            for i in tqdm(range(len(wno)), desc=species):
                arr = full_opac[:, :, i]
                _, beta = fit_mlr(arr, xMat)
                diff = (arr - np.dot(xMat, beta).T) / arr
                species_means += [np.median(np.abs(diff))]
                betas += [beta]

            species_means_dict[species] = np.array(species_means)

#         np.save(f'{species}_vectors.npy', xMat)
#         np.save(f'{species}_pca_coeffs.npy', np.array(betas))



"""
this file holds the class describing opacity data. hmm...maybe I want the loader in a different file?

author: @arjunsavel
"""
from cortecs.opac.io import *


class Opac(object):
    """
    this class holds the opacity data and provides methods for evaluating the opacity at a given temperature and pressure.

    Everything's already been loaded into memory, so this is just a wrapper for the data.

    todo: use the different loaders
    """

    method_dict = {
        "chimera": loader_base,
        "helios": loader_helios,
        "platon": loader_platon,
        "exotransmit": loader_exotransmit,
    }

    wl = None
    T = None
    P = None

    def __init__(self, filename, loader="chimera", load_kwargs={}):
        """
        wraps around the loaders.

        Parameters
        ----------
        filename : str
            name of file to load
        loader : str
            name of loader to use. default is chimera.

        Returns
        -------
        nothing
        """
        self.filename = filename
        self.load_obj = self.method_dict[loader]()
        self.wl, self.T, self.P, self.cross_section = self.load_obj.load(
            filename, **load_kwargs
        )

        self.n_wav, self.n_t, self.n_p = len(self.wl), len(self.T), len(self.P)

        return

    def _get_loader(self, loader_name):
        """
        gets the loader object

        Parameters
        ----------
        loader_name : str
            name of loader

        Returns
        -------
        loader object
        """
        if loader_name not in self.method_dict:
            raise ValueError(
                "loader name not found. valid loaders are: {}".format(
                    self.method_dict.keys()
                )
            )
        return self.method_dict[loader_name]


class Opac_cia(Opac):
    """
    this class holds the opacity data and provides methods for evaluating the opacity at a given temperature and pressure.

    Everything's already been loaded into memory, so this is just a wrapper for the data.

    todo: add exotransmit CIA
    """

    method_dict = {
        "platon_cia": loader_platon_cia,
    }

    def __init__(self, filename, loader="chimera"):
        """
        wraps around the loaders.

        Parameters
        ----------
        filename : str
            name of file to load
        loader : str
            name of loader to use. default is chimera.

        Returns
        -------
        nothing
        """
        self.filename = filename
        load_obj = loader_base()
        self.wl, self.T, self.cross_section = load_obj.load(filename, loader=loader)

        self.n_wav, self.n_t = len(self.wl), len(self.T)

        return



import jax
import numpy as np


@jax.jit
def eval_pca_ind_wav(temperature_ind, pressure_ind, vectors, pca_coeffs):
    """
    Evaluates the PCA fit at a given temperature and pressure.

    Unfortunately, not all GPUs will support a simpler dot product, I believe,
    Also, we cannot loop over n_components explicitly because JAX
    functions require static loop lengths.

    Inputs
    ------
    temperature_ind: int
        The index of the temperature to evaluate at.
    pressure_ind: int
        The index of the pressure to evaluate at.
    vectors: array
        The PCA vectors.
    pca_coeffs: array
        The PCA coefficients.
    n_components: int
        The number of PCA components used in the fitting


    """

    xsec_val = 0.0
    n_components = vectors.shape[1]
    for component in range(n_components):
        xsec_val += (
            vectors[temperature_ind, component] * pca_coeffs[component, pressure_ind]
        )
    return xsec_val


def eval_pca(temperature, pressure, wavelength, T, P, wl, fitter_results):
    """
    Evaluates the PCA fit at a given temperature, pressure, and wavelength.


    Inputs
    ------
    temperature: float
        The temperature to evaluate at.
    pressure: float
        The pressure to evaluate at.
    wavelength: float
        The wavelength to evaluate at.

    """
    temperature_ind = np.where(np.isclose(T, temperature))[0][0]
    pressure_ind = np.where(np.isclose(P, pressure))[0][0]
    wavelength_ind = np.where(np.isclose(wl, wavelength))[0][0]
    pca_vectors, pca_coeffs_all_wl = fitter_results
    pca_coeffs = pca_coeffs_all_wl[wavelength_ind, :, :]

    return eval_pca_ind_wav(temperature_ind, pressure_ind, pca_vectors, pca_coeffs)
